import functools
import logging
from contextlib import asynccontextmanager
from copy import deepcopy
from importlib.metadata import entry_points
from inspect import iscoroutinefunction, signature
from typing import Callable, final

from pydantic import BaseModel, field_validator, ConfigDict
from typing_extensions import Self

from icij_common.import_utils import VariableNotFound, import_variable
from icij_common.pydantic_utils import ICIJSettings, icij_config
from icij_worker.routing_strategy import RoutingStrategy
from icij_worker.typing_ import Dependency
from icij_worker.utils import run_deps

logger = logging.getLogger(__name__)

PROGRESS_HANDLER_ARG = "progress"


class TaskGroup(BaseModel):
    model_config = icij_config()

    name: str
    timeout_s: int | None = None
    max_task_queue_size: int | None = None


class AsyncAppConfig(ICIJSettings):
    late_ack: bool = True
    recover_from_worker_timeout: bool = False
    max_task_queue_size: int | None = None
    model_config = ConfigDict(env_prefix="ICIJ_APP_")


class RegisteredTask(BaseModel):
    model_config = icij_config()

    task: Callable
    recover_from: tuple[type[Exception], ...] = tuple()
    max_retries: int | None
    group: TaskGroup | None
    timeout_s: int | None

    @field_validator("group", mode="before")
    @classmethod
    def validate_group_instance(cls, v):  # pylint: disable=no-self-argument
        if isinstance(v, str):
            v = TaskGroup(name=v)
        return v


class AsyncApp:
    def __init__(
        self,
        name: str,
        config: AsyncAppConfig = None,
        dependencies: list[Dependency] | None = None,
        routing_strategy: RoutingStrategy | None = None,
    ):
        self._name = name
        if config is None:
            config = AsyncAppConfig()  # This will load from the env
        self._config = config
        self._registry = dict()
        self._groups = dict()
        if dependencies is None:
            dependencies = []
        self._dependencies = dependencies
        if routing_strategy is None:
            routing_strategy = RoutingStrategy()
        self._routing_strategy = routing_strategy

    @property
    def config(self) -> AsyncAppConfig:
        return self._config

    def with_config(self, value: AsyncAppConfig) -> Self:
        if not isinstance(value, AsyncAppConfig):
            raise TypeError(f"Expected {AsyncAppConfig.__name__}, got {value}")
        self._config = value
        return self

    @property
    def registry(self) -> dict[str, RegisteredTask]:
        return self._registry

    @property
    def registered_keys(self) -> list[str]:
        return sorted(self._registry)

    @functools.cached_property
    def name(self) -> str:
        return self._name

    @functools.cached_property
    def routing_strategy(self) -> RoutingStrategy:
        return self._routing_strategy

    def with_routing_strategy(self, ns: RoutingStrategy) -> Self:
        self._routing_strategy = ns
        return self

    def task(
        self,
        name: str | None = None,
        recover_from: tuple[type[Exception]] = tuple(),
        max_retries: int | None = None,
        *,
        group: str | TaskGroup | None = None,
        timeout_s: int | None = None,
    ) -> Callable:
        if callable(name) and not recover_from and max_retries is None:
            f = name
            return functools.partial(
                self._register_task, name=f.__name__, group=group, timeout_s=None
            )(f)
        if max_retries is None:
            max_retries = 3
        return functools.partial(
            self._register_task,
            name=name,
            recover_from=recover_from,
            max_retries=max_retries,
            group=group,
            timeout_s=timeout_s,
        )

    def task_group(self, name: str) -> TaskGroup | None:
        return self._groups.get(name)

    @property
    def task_groups(self) -> list[TaskGroup]:
        return list(self._groups.values())

    @final
    @asynccontextmanager
    async def lifetime_dependencies(self, **kwargs):
        ctx = f"{self.name} async app"
        async with run_deps(self._dependencies, ctx=ctx, **kwargs):
            yield

    def _register_task(
        self,
        f: Callable,
        *,
        name: str | None = None,
        recover_from: tuple[type[Exception]] = tuple(),
        max_retries: int | None = None,
        group: str | TaskGroup | None = None,
        timeout_s: int | None,
    ) -> Callable:
        if not iscoroutinefunction(f) and supports_progress(f):
            msg = (
                f"{f} is not a coroutine, progress is not supported as progress"
                f" reporting is inherently async, turn your function task into a"
                f" coroutine if necessary and use `await progress(my_progress)`"
            )
            raise ValueError(msg)
        if name is None:
            name = f.__name__
        registered = self._registry.get(name)
        if registered is not None:
            raise ValueError(f'Task "{name}" is already registered: {registered}')
        registered = RegisteredTask(
            task=f,
            max_retries=max_retries,
            recover_from=recover_from,
            group=group,
            timeout_s=timeout_s,
        )
        self._validate_group(registered)
        self._registry[name] = registered
        if registered.group is not None:
            self._groups[registered.group.name] = registered.group

        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            return f(*args, **kwargs)

        return wrapped

    def _validate_group(self, task: RegisteredTask):
        if task.group is None:
            return
        existing = self._groups.get(task.group.name)
        if existing is not None and existing.name != task.group.name:
            msg = (
                f"invalid task group {task.group}, it has the same name as registered "
                f"group {existing}, use {existing} directly or specify a different name"
            )
            raise ValueError(msg)

    @classmethod
    def load(cls, app_path: str, config: AsyncAppConfig | None = None) -> Self:
        try:
            app = import_variable(app_path)
        except VariableNotFound as e:
            app_plugins = entry_points(group="icij_worker.APP_HOOK")
            for entry_point in app_plugins:
                if entry_point.name == app_path:
                    app = entry_point.load()
                    break
            else:
                msg = (
                    f"invalid app path {app_path}, not found in available modules"
                    f" nor in icij_worker plugins"
                )
                raise ValueError(msg) from e
        app = deepcopy(app)
        if config is not None:
            app.with_config(config)
        return app

    def filter_tasks(self, group: str | None) -> Self:
        if group is None:
            return self
        kept = {
            t_name
            for t_name, t in self._registry.items()
            if self._routing_strategy.app_tasks_filter(
                task_group=t.group, app_group_name=group
            )
        }
        discarded = set(self._registry) - kept
        logger.info(
            "Applied group filtering:\n- running: %s\n- discarded: %s",
            ", ".join(sorted(kept)),
            ", ".join(sorted(discarded)),
        )
        self._registry = {k: self._registry[k] for k in kept}
        return self

    def __deepcopy__(self, memodict={}) -> Self:
        # pylint: disable=dangerous-default-value
        app = AsyncApp(
            name=self.name,
            config=deepcopy(self.config),
            dependencies=list(self._dependencies),
            routing_strategy=self.routing_strategy,
        )
        app._registry = deepcopy(self._registry)
        app._groups = deepcopy(self._groups)
        return app


def supports_progress(task_fn) -> bool:
    return any(
        param.name == PROGRESS_HANDLER_ARG
        for param in signature(task_fn).parameters.values()
    )



from __future__ import annotations

import logging
from datetime import datetime, timezone
from functools import cached_property
from typing import TypeVar, cast

from aio_pika import connect_robust
from aio_pika.abc import AbstractExchange, AbstractQueueIterator
from aiormq import DeliveryError
from pydantic import Field

from icij_worker.app import AsyncApp
from icij_worker.exceptions import MessageDeserializationError, TaskQueueIsFull
from icij_worker.objects import (
    AsyncBackend,
    CancelEvent,
    ErrorEvent,
    ManagerEvent,
    Message,
    ResultEvent,
    ShutdownEvent,
    Task,
    TaskState,
)
from icij_worker.routing_strategy import Routing
from icij_worker.task_manager import TaskManager, TaskManagerConfig
from icij_worker.task_storage import TaskStorage
from icij_worker.task_storage.fs import FSKeyValueStorageConfig
from icij_worker.task_storage.postgres import PostgresStorageConfig
from icij_worker.utils.amqp import (
    AMQPConfigMixin,
    AMQPManagementClient,
    AMQPMixin,
    RobustConnection,
    amqp_task_group_policy,
    health_policy,
)

logger = logging.getLogger(__name__)

S = TypeVar("S", bound=TaskStorage)


@TaskManagerConfig.register()
class AMQPTaskManagerConfig(TaskManagerConfig, AMQPConfigMixin):
    backend: AsyncBackend = Field(frozen=True, default=AsyncBackend.amqp)
    storage: FSKeyValueStorageConfig | PostgresStorageConfig


@TaskManager.register(AsyncBackend.amqp)
class AMQPTaskManager(TaskManager, AMQPMixin):
    def __init__(
        self,
        app: AsyncApp,
        task_store: TaskStorage,
        management_client: AMQPManagementClient,
        *,
        broker_url: str,
        connection_timeout_s: float = 1.0,
        reconnection_wait_s: float = 5.0,
        inactive_after_s: float | None = None,
        is_qpid: bool = False,
    ):
        super().__init__(app)
        super(TaskManager, self).__init__(  # pylint: disable=bad-super-call
            broker_url,
            connection_timeout_s=connection_timeout_s,
            reconnection_wait_s=reconnection_wait_s,
            inactive_after_s=inactive_after_s,
            is_qpid=is_qpid,
        )
        self._management_client = management_client
        self._storage = task_store
        self._task_x: AbstractExchange | None = None
        self._worker_evt_x: AbstractExchange | None = None
        self._health_x: AbstractExchange | None = None
        self._manager_messages_it: AbstractQueueIterator | None = None

        self._task_groups: dict[str, str | None] = dict()

    @classmethod
    def _from_config(cls, config: AMQPTaskManagerConfig, **extras) -> AMQPTaskManager:
        # pylint: disable=unused-argument
        app = config.app
        storage = config.storage.to_storage(app.routing_strategy)
        management_client = config.to_management_client()
        task_manager = cls(
            app,
            storage,
            management_client,
            broker_url=config.broker_url,
            connection_timeout_s=config.connection_timeout_s,
            reconnection_wait_s=config.reconnection_wait_s,
            is_qpid=config.rabbitmq_is_qpid,
        )
        return task_manager

    async def _aenter__(self) -> AMQPTaskManager:
        logger.info("starting task manager connection workflow...")
        await self._exit_stack.__aenter__()  # pylint: disable=unnecessary-dunder-call
        await self._exit_stack.enter_async_context(self._storage)
        await self._connection_workflow()
        self._manager_messages_it = (
            await self._get_queue_iterator(
                self.manager_evt_routing(),
                declare_exchanges=False,
                declare_queues=False,
            )
        )[0]
        return self

    async def _aexit__(self, exc_type, exc_val, exc_tb):
        await self._exit_stack.__aexit__(exc_type, exc_val, exc_tb)

    @cached_property
    def _app_id(self) -> str:
        return self.app_name

    async def get_task(self, task_id: str) -> Task:
        return await self._storage.get_task(task_id)

    async def get_task_group(self, task_id: str) -> str | None:
        return await self._storage.get_task_group(task_id)

    async def get_task_errors(self, task_id: str) -> list[ErrorEvent]:
        return await self._storage.get_task_errors(task_id)

    async def get_task_result(self, task_id: str) -> ResultEvent:
        return await self._storage.get_task_result(task_id)

    async def get_tasks(
        self,
        group: str | None,
        *,
        task_name: str | None = None,
        state: list[TaskState] | TaskState | None = None,
        **kwargs,
    ) -> list[Task]:
        # pylint: disable=unused-argument
        return await self._storage.get_tasks(group, task_name=task_name, state=state)

    async def save_task_(self, task: Task, group: str | None) -> bool:
        return await self._storage.save_task_(task, group)

    async def save_result(self, result: ResultEvent):
        await self._storage.save_result(result)

    async def save_error(self, error: ErrorEvent):
        await self._storage.save_error(error)

    async def _consume(self) -> ManagerEvent:
        # pylint: disable=unnecessary-dunder-call
        msg = await self._manager_messages_it.__anext__()
        try:
            message = cast(ManagerEvent, Message.model_validate_json(msg.body))
        except Exception as e:
            msg = f"invalid manager event body {msg.body}"
            raise MessageDeserializationError(msg) from e
        await msg.ack()
        return message

    async def _enqueue(self, task: Task):
        group = await self._storage.get_task_group(task.id)
        routing = self._routing_strategy.amqp_task_routing(group)
        try:
            await self._publish_message(
                task,
                exchange=self._task_x,
                routing_key=routing.routing_key,
                mandatory=True,  # This is important
            )
        except DeliveryError as e:
            raise TaskQueueIsFull(self.max_task_queue_size) from e

    async def _requeue(self, task: Task):
        group = await self._storage.get_task_group(task.id)
        routing = self._routing_strategy.amqp_task_routing(group)
        try:
            await self._publish_message(
                task,
                exchange=self._task_x,
                routing_key=routing.routing_key,
                mandatory=True,  # This is important
            )
        except DeliveryError as e:
            raise TaskQueueIsFull(self.max_task_queue_size) from e

    async def cancel(self, task_id: str, *, requeue: bool):
        cancel_event = CancelEvent(
            task_id=task_id, requeue=requeue, created_at=datetime.now(timezone.utc)
        )
        # TODO: for now cancellation is not grouped, workers from other group
        #  are responsible to ignoring the broadcast. That could be easily implemented
        #  in the future but will need sync with Java
        routing = self.worker_evt_routing().routing_key
        await self._publish_message(
            cancel_event,
            exchange=self._worker_evt_x,
            routing_key=routing,
            mandatory=True,  # This is important
        )

    async def shutdown_workers(self):
        shutdown_event = ShutdownEvent(created_at=datetime.now(timezone.utc))
        # TODO: for now cancellation is not grouped, workers from other group
        #  are responsible to ignoring the broadcast. That could be easily implemented
        #  in the future but will need sync with Java
        routing = self.worker_evt_routing().routing_key
        await self._publish_message(
            shutdown_event,
            exchange=self._worker_evt_x,
            routing_key=routing,
            mandatory=True,  # This is important
        )

    async def _connection_workflow(self):
        await self._exit_stack.enter_async_context(self._management_client)
        logger.debug("creating connection...")
        self._connection_ = await connect_robust(
            self._broker_url,
            timeout=self._connection_timeout_s,
            reconnect_interval=self._reconnection_wait_s,
            connection_class=RobustConnection,
        )
        await self._exit_stack.enter_async_context(self._connection)
        logger.debug("creating channel...")
        self._channel_ = await self._connection.channel(
            publisher_confirms=self._publisher_confirms, on_return_raises=False
        )
        await self._exit_stack.enter_async_context(self._channel)
        await self._channel.set_qos(prefetch_count=1, global_=False)
        logger.info("channel opened !")
        logger.info("creating task queues opened !")
        await self._ensure_task_queues()
        for routing in self._other_routings:
            logger.debug("(re)declaring routing %s...", routing)
            await self._create_routing(
                routing, declare_exchanges=True, declare_queues=True
            )
        await self._create_routing(
            self.worker_evt_routing(), declare_exchanges=True, declare_queues=False
        )
        await self._create_routing(
            self.manager_evt_routing(), declare_exchanges=True, declare_queues=True
        )
        await self._create_routing(
            self.health_routing(), declare_exchanges=True, declare_queues=True
        )
        self._task_x = await self._channel.get_exchange(
            self.default_task_routing().exchange.name, ensure=True
        )
        self._manager_evt_x = await self._channel.get_exchange(
            self.manager_evt_routing().exchange.name, ensure=True
        )
        self._worker_evt_x = await self._channel.get_exchange(
            self.worker_evt_routing().exchange.name, ensure=True
        )
        self._health_x = await self._channel.get_exchange(
            self.health_routing().exchange.name, ensure=True
        )
        healthz_policy = health_policy(self.health_routing())
        await self._management_client.set_policy(healthz_policy)
        logger.info("connection workflow complete")

    @cached_property
    def _other_routings(self) -> list[Routing]:
        worker_events_routing = AMQPMixin.worker_evt_routing()
        manager_events_routing = AMQPMixin.manager_evt_routing()
        return [worker_events_routing, manager_events_routing]

    async def _ensure_task_queues(self):
        default_routing = AMQPMixin.default_task_routing()
        group_policy = amqp_task_group_policy(
            default_routing, None, self.max_task_queue_size
        )
        await self._create_routing(
            default_routing, declare_exchanges=True, declare_queues=True
        )
        await self._management_client.set_policy(group_policy)
        for group in self._app.task_groups:
            routing = self._routing_strategy.amqp_task_routing(group.name)
            group_policy = amqp_task_group_policy(
                routing, group, self.max_task_queue_size
            )
            await self._management_client.set_policy(group_policy)
            await self._create_routing(
                routing, declare_exchanges=True, declare_queues=True
            )

    async def get_health(self) -> dict[str, bool]:
        storage_health = await self._storage.get_health()
        amqp_health = await self._get_amqp_health()
        return {"storage": storage_health, "amqp": amqp_health}



from abc import ABC, abstractmethod
from typing import final

from icij_worker import ManagerEvent


class EventPublisher(ABC):
    @final
    async def publish_event(self, event: ManagerEvent):
        await self._publish_event(event)

    @abstractmethod
    async def _publish_event(self, event: ManagerEvent):
        pass



from types import TracebackType
from typing import Any, Awaitable, Callable, Coroutine, Mapping, Protocol

from typing_extensions import AbstractSet

DependencyLabel = str | None
DependencySetup = Callable[..., None]
DependencyAsyncSetup = Callable[..., Coroutine[None, None, None]]

RateProgress = Callable[[float], Awaitable[None]]
# TODO: remove this when breaking API
PercentProgress = RateProgress
RawProgress = Callable[[int], Awaitable[None]]

DictStrAny = dict[str, Any]
IntStr = int | str
AbstractSetIntStr = AbstractSet[IntStr]
MappingIntStrAny = Mapping[IntStr, Any]


class DependencyTeardown(Protocol):
    def __call__(
        self,
        exc_type: type[Exception] | None,
        exc_value: Exception | None,
        traceback: TracebackType | None,
    ) -> None: ...


class DependencyAsyncTeardown(Protocol):
    async def __call__(
        self,
        exc_type: type[Exception] | None,
        exc_value: Exception | None,
        traceback: TracebackType | None,
    ) -> None: ...


Dependency = tuple[
    DependencyLabel,
    DependencySetup | DependencyAsyncSetup,
    DependencyTeardown | DependencyAsyncTeardown | None,
]



import asyncio
from abc import ABC
from typing import Awaitable, Callable, TypeVar

import neo4j

from icij_common.neo4j_.migrate import retrieve_dbs
from icij_worker.objects import TaskEvent
from icij_worker.task_storage.neo4j_ import Neo4jStorage

T = TypeVar("T", bound=TaskEvent)
ConsumeT = Callable[[neo4j.AsyncTransaction, ...], Awaitable[T | None]]


class Neo4jConsumerMixin(Neo4jStorage, ABC):
    _driver: neo4j.AsyncDriver
    _db_driver: neo4j.AsyncDriver
    _task_meta: dict[str, tuple[str, str]] = dict()

    async def _consume_(
        self,
        consume_tx: ConsumeT,
        refresh_interval_s: float,
        db_filter: Callable[[str], bool] | None,
    ) -> T:
        dbs = []
        refresh_dbs_i = 0
        while "i'm waiting until I find something interesting":
            # Refresh project list once in an while
            refresh_dbs = (refresh_dbs_i % 10) == 0
            if refresh_dbs:
                dbs = await retrieve_dbs(self._driver)
                if db_filter is not None:
                    dbs = [db for db in dbs if db_filter(db.name)]
            for db in dbs:
                async with self._db_session(db.name) as sess:
                    received = await sess.execute_write(consume_tx)
                    if "group" in received:
                        self._task_meta[received.id] = (db.name, received["group"])
                    if received is not None:
                        return received
            await asyncio.sleep(refresh_interval_s)
            refresh_dbs_i += 1



import logging
from contextlib import contextmanager
from enum import Enum
from pathlib import Path

from icij_worker import WorkerConfig
from icij_worker.backend.mp import (
    run_workers_with_multiprocessing,
    run_workers_with_multiprocessing_cm,
)

logger = logging.getLogger(__name__)


class WorkerBackend(str, Enum):
    # pylint: disable=invalid-name

    # We could support more backend type, and for instance support asyncio/thread backed
    # workers for IO based tasks
    MULTIPROCESSING = "multiprocessing"

    def run(
        self,
        app: str,
        *,
        n_workers: int,
        config: WorkerConfig,
        worker_extras: dict | None = None,
        app_deps_extras: dict | None = None,
        group: str | None,
    ):
        run_workers_with_multiprocessing(
            app,
            n_workers,
            config,
            worker_extras=worker_extras,
            app_deps_extras=app_deps_extras,
            group=group,
        )

    # TODO: remove this when the HTTP server doesn't
    # TODO: also refactor underlying functions to be simple function rather than
    #  context managers
    @contextmanager
    def run_cm(
        self,
        app: str,
        *,
        n_workers: int,
        config: WorkerConfig,
        worker_extras: dict | None = None,
        app_deps_extras: dict | None = None,
        group: str | None,
    ):
        if self is WorkerBackend.MULTIPROCESSING:
            with run_workers_with_multiprocessing_cm(
                app,
                n_workers,
                config,
                worker_extras=worker_extras,
                app_deps_extras=app_deps_extras,
                group=group,
            ):
                yield
        else:
            raise NotImplementedError(f"Can't start workers with backend: {self}")


def start_workers(
    app: str,
    n_workers: int,
    config_path: Path | None,
    backend: WorkerBackend,
    group: str | None,
):
    if n_workers < 1:
        raise ValueError("n_workers must be >= 1")
    if config_path is not None:
        logger.info("Loading worker configuration from %s", config_path)
        config = WorkerConfig.model_validate_json(config_path.read_text())
    else:
        logger.info("Loading worker configuration from env...")
        config = WorkerConfig.from_env()
    logger.info("worker config: %s", config.model_dump_json(indent=2))

    backend.run(app, n_workers=n_workers, config=config, group=group)
